from supervisely.nn.benchmark.visualization.widgets.chart.chart import ChartWidget
from supervisely.nn.benchmark.visualization.widgets.collapse.collapse import (
    CollapseWidget,
)
from supervisely.nn.benchmark.visualization.widgets.container.container import (
    ContainerWidget,
)
from supervisely.nn.benchmark.visualization.widgets.gallery.gallery import GalleryWidget
from supervisely.nn.benchmark.visualization.widgets.markdown.markdown import (
    MarkdownWidget,
)
from supervisely.nn.benchmark.visualization.widgets.notification.notification import (
    NotificationWidget,
)
from supervisely.nn.benchmark.visualization.widgets.sidebar.sidebar import SidebarWidget
from supervisely.nn.benchmark.visualization.widgets.table.table import TableWidget
from supervisely.nn.benchmark.visualization.widgets.widget import BaseWidget
